"""
Fields for basic data types.
"""

from bson import ObjectId

from yadm.fields.base import Field
from yadm.markers import NoDefault


class SimpleField(Field):
    """ Base field for simple types

    :param default: default value
    :param set choices: set of possible values
    """
    type = NotImplemented
    choices = None

    def __init__(self, default=NoDefault, choices=None):
        if self.type is NotImplemented:
            raise ValueError('Attribute "type" not implemented!')

        self.choices = choices

        super().__init__(default)

    def prepare_value(self, document, value):
        if not isinstance(value, self.type) and value is not None:
            value = self.type(value)

        if self.choices is not None and value not in self.choices:
            raise ValueError('value not in choices: {!r}'.format(value))

        return value

    def to_mongo(self, document, value):
        return self.prepare_value(document, value)

    def from_mongo(self, document, value):
        return self.prepare_value(document, value)


class ObjectIdField(SimpleField):
    """ Field for ObjectId

    :param bool default_gen: generate default value if not set
    """
    type = ObjectId

    def __init__(self, default_gen=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_gen = default_gen

    @property
    def default(self):
        if self.default_gen:
            return ObjectId()
        else:
            return NoDefault


class BooleanField(SimpleField):
    """ Field for boolean values
    """
    type = bool


class IntegerField(SimpleField):
    """ Field for integer
    """
    type = int


class FloatField(SimpleField):
    """ Field for float
    """
    type = float


class StringField(SimpleField):
    """ Field for string
    """
    type = str



from unittest import TestCase

from bson import ObjectId

from yadm.documents import Document
from yadm.fields.simple import StringField, IntegerField


class ChoicesTest(TestCase):
    def setUp(self):
        class TestDoc(Document):
            string = StringField(choices={'qwerty', 'zzz', 'asd'})

        self.TestDoc = TestDoc

    def test_set_valid(self):
        doc = self.TestDoc()
        doc.string = 'zzz'
        self.assertEqual(doc.string, 'zzz')

    def test_set_invalid(self):
        doc = self.TestDoc()
        self.assertRaises(ValueError, setattr, doc, 'string', 'invalid')


class ChoicesWithDefaultTest(TestCase):
    def test_valid(self):
        class TestDoc(Document):
            string = IntegerField(default=0, choices={0, 13, 42})

        doc = TestDoc()
        self.assertEqual(doc.string, 0)

    def test_invalid(self):
        with self.assertRaises(ValueError):
            IntegerField(default=1, choices={0, 13, 42})


class ObjectIdFieldTest(TestCase):
    def setUp(self):
        class TestDoc(Document):
            pass

        self.TestDoc = TestDoc

    def test_default(self):
        self.assertRaises(AttributeError, getattr, self.TestDoc(), '_id')

    def test_setattr_objectid(self):
        _id = ObjectId()
        td = self.TestDoc()
        td._id = _id

        self.assertIsInstance(td._id, ObjectId)
        self.assertEqual(td._id, _id)

    def test_setattr_str(self):
        _id = ObjectId()
        td = self.TestDoc()
        td._id = str(_id)

        self.assertIsInstance(td._id, ObjectId)
        self.assertEqual(td._id, _id)



from datetime import datetime, timezone

from yadm import fields
from yadm.documents import Document

from .test_database import BaseDatabaseTest


class DatetimeFieldTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class TestDoc(Document):
            __collection__ = 'testdocs'
            dt = fields.DatetimeField()
            now = fields.DatetimeField(auto_now=True)

        self.TestDoc = TestDoc

    def test_cast(self):
        doc = self.TestDoc()
        doc.dt = datetime(1970, 1, 1, tzinfo=timezone.utc).isoformat()
        self.assertIsInstance(doc.dt, datetime)

    def test_default_none(self):
        doc = self.TestDoc()
        self.assertFalse(hasattr(doc, 'dt'))

    def test_save(self):
        doc = self.TestDoc()
        doc.dt = datetime(1970, 1, 1, tzinfo=timezone.utc)
        self.db.insert(doc)

        doc = self.db.get_queryset(self.TestDoc).with_id(doc.id)
        data = self.db.db.testdocs.find_one()

        self.assertIsInstance(data['dt'], datetime)
        self.assertEqual(data['dt'], doc.dt)

    def test_load(self):
        epoch = datetime(1970, 1, 1, tzinfo=timezone.utc)
        _id = self.db.db.testdocs.insert({'dt': epoch})

        doc = self.db.get_queryset(self.TestDoc).with_id(_id)

        self.assertIsInstance(doc.dt, datetime)
        self.assertEqual(doc.dt, epoch)

    def test_default_auto_now(self):
        doc = self.TestDoc()
        self.assertTrue(hasattr(doc, 'now'))
        self.assertIsInstance(doc.now, datetime)

    def test_default_now_save(self):
        doc = self.TestDoc()
        self.db.insert(doc)

        doc = self.db.get_queryset(self.TestDoc).with_id(doc.id)
        data = self.db.db.testdocs.find_one()

        self.assertIsInstance(data['now'], datetime)
        self.assertEqual(data['now'], doc.now)

    def test_now_save(self):
        epoch = datetime(1970, 1, 1, tzinfo=timezone.utc)

        doc = self.TestDoc()
        doc.now = epoch
        self.db.insert(doc)

        doc = self.db.get_queryset(self.TestDoc).with_id(doc.id)
        data = self.db.db.testdocs.find_one()

        self.assertIsInstance(data['now'], datetime)
        self.assertEqual(data['now'], epoch)

    def test_now_load(self):
        epoch = datetime(1970, 1, 1, tzinfo=timezone.utc)
        _id = self.db.db.testdocs.insert({'now': epoch})

        doc = self.db.get_queryset(self.TestDoc).with_id(_id)

        self.assertIsInstance(doc.now, datetime)
        self.assertEqual(doc.now, epoch)



from yadm.serialize import from_mongo


class QuerySet:
    def __init__(self, db, document_class):
        self._db = db
        self._document_class = document_class
        self._criteria = None
        self._projection = None
        self._sort = []

    def __str__(self):
        return ('QuerySet({s._document_class.__collection__},'
                ' {s._criteria!r}, {s._projection!r}, {s._sort!r})'.format(s=self))

    def __len__(self):
        return self.count()

    def __iter__(self):
        return self._from_mongo_list(self._cursor)

    def __contains__(self, document):
        return self.with_id(document.id) is not None

    def __getitem__(self, item):
        if isinstance(item, slice):
            return self._from_mongo_list(self._cursor[item])

        elif isinstance(item, int):
            return self._from_mongo_one(self._cursor[item])

        else:
            raise TypeError('Only slice or int accepted')

    def __call__(self, criteria=None, projection=None):
        return self.find(criteria, projection)

    def _from_mongo_one(self, data):
        """ Create document from raw data
        """
        if data is not None:
            doc = from_mongo(self._document_class, data)
            doc.__db__ = self._db
            return doc
        else:
            return None

    def _from_mongo_list(self, data):
        """ Generator for got documents from raw data list (cursor)
        """
        for d in data:
            yield self._from_mongo_one(d)

    @property
    def _collection(self):
        """ pymongo collection
        """
        return self._db._get_collection(self._document_class)

    @property
    def _cursor(self):
        """ pymongo cursor with parameters from queryset
        """
        cursor = self._collection.find(self._criteria, self._projection or None)

        if self._sort:
            cursor = cursor.sort(self._sort)

        return cursor

    def _copy_qs(self):
        """ Copy queryset instance
        """
        qs = self.__class__(self._db, self._document_class)

        if self._criteria is not None:
            qs._criteria = self._criteria.copy()

        if self._projection is not None:
            qs._projection = self._projection.copy()

        qs._sort = self._sort[:]

        return qs

    def _update_qs(self, **kwargs):
        """ Update queryset parameters in place
        """
        if 'criteria' in kwargs:
            if self._criteria is None:
                self._criteria = kwargs['criteria']
            else:
                self._criteria.update(kwargs['criteria'] or {})

        if 'projection' in kwargs:
            if self._projection is None:
                self._projection = kwargs['projection']
            else:
                self._projection.update(kwargs['projection'] or {})

        if 'sort' in kwargs and kwargs['sort'] is not None:
            self._sort.extend(kwargs['sort'])

    def copy(self, **kwargs):
        """ Copy queryset and update it

            :criteria:
            :projection:
            :sort:
        """
        qs = self._copy_qs()
        qs._update_qs(**kwargs)
        return qs

    def find(self, criteria=None, projection=None):
        """ Return queryset copy with new criteria and projection

            qs({'field': {'$gt': 3}}, {'field': True})
        """
        return self.copy(criteria=criteria, projection=projection)

    def find_one(self, criteria=None, projection=None):
        """ Find and return only one document

            qs({'field': {'$gt': 3}}, {'field': True})
        """
        qs = self.copy(criteria=criteria, projection=projection)
        collection = qs._db._get_collection(qs._document_class)
        data = collection.find_one(qs._criteria, qs._projection)
        return self._from_mongo_one(data)

    def with_id(self, _id):
        """ Find document with id
        """
        doc = self._document_class()
        doc._id = _id
        return self.find_one({'_id': doc._id})

    def update(self, update, multi=True):
        """ Update documents in queryset
        """
        return self._collection.update(
            self._criteria,
            update,
            multi=multi,
            upsert=False,
        )

    def find_and_modify(
            self, update=None, upsert=False,
            full_response=False, new=False,
            **kwargs):
        """ Execute findAndModify query

        :param dict update: see second argument to update()
        :param bool upsert: insert if object doesn’t exist
            (default False)
        :param bool full_response: return the entire response
            object from the server (default False)
        :param new: return updated rather than original object
            (default False)
        :param **kwargs: any other options the findAndModify
            command supports can be passed here
        """
        result = self._collection.find_and_modify(
            query=self._criteria,
            update=update,
            upsert=upsert,
            sort=self._sort or None,
            full_response=full_response,
            new=new,
            **kwargs
        )
        if not full_response:
            return self._from_mongo_one(result)
        else:
            result['value'] = self._from_mongo_one(result['value'])
            return result

    def remove(self):
        """ Remove documents from queryset
        """
        return self._collection.remove(self._criteria)

    def fields(self, *fields):
        """ Get only setted fields

        Update projection with fields.

            qs(field, field2)
        """
        return self.copy(projection=dict.fromkeys(fields, True))

    def fields_all(self):
        """ Clear projection
        """
        qs = self.copy()
        qs._projection = None
        return qs

    def sort(self, *sort):
        """
            qs.sort(('field_1', 1), ('field_2', -1))
        """
        return self.copy(sort=sort)

    def count(self):
        """ Count documents in queryset
        """
        return self._cursor.count()

    def bulk(self):
        """ Return map {id: object}
        """
        qs = self.copy()
        qs._sort = []
        return {obj.id: obj for obj in qs}



"""
Field for money

Work with `decimal.Decimal` and store value as integer.
Use :class:`yadm.fields.money.Money`, as value for money.

.. code block: python

    class DocClass(Document):
        money = MoneyField()

    doc = DocClass()
    doc.money = Money('3.14')

    db.insert(doc)

This code save to MongoDB document:

.. code block: javascript

    {
        id: ObjectId('534272984c78591787e1a964'),
        money: 314
    }


"""
from decimal import Decimal, Context, ROUND_UP

from yadm.fields.base import Field


class Money(Decimal):
    context = Context(rounding=ROUND_UP)

    def __new__(cls, value):
        return Decimal.__new__(cls, value, cls.context)

    def to_mongo(self):
        return int((self * 100).quantize(Decimal('1.00'), ROUND_UP))

    def __str__(self):
        s = str(self.to_mongo()).rjust(3, '0')
        return '.'.join((s[:-2], s[-2:]))

    def __xor__(self, other):
        return NotImplemented

    def __rxor__(self, other):
        return NotImplemented

    def __mod__(self, other):
        return NotImplemented

    def __rmod__(self, other):
        return NotImplemented

    def __divmod__(self, other):
        return NotImplemented

    def __rdivmod__(self, other):
        return NotImplemented


class MoneyField(Field):
    """ Field for work with money
    """
    def prepare_value(self, document, value):
        """ Cast value to :class:`decimal.Decimal`
        """
        if value is None:
            return None

        elif isinstance(value, Money):
            return value

        elif isinstance(value, (str, Decimal)):
            return Money(value)

        else:
            raise TypeError(repr(value))

    def to_mongo(self, document, value):
        if value is None:
            return None

        elif isinstance(value, (str, Decimal)):
            return Money(value).to_mongo()

        elif isinstance(value, Money):
            return value.to_mongo()

        elif isinstance(value, int):
            return value

        else:
            raise TypeError(repr(value))

    def from_mongo(self, document, data):
        if isinstance(data, int):
            return Money(data / Decimal(100))
        else:
            return self.prepare_value(document, data)



from unittest import TestCase, SkipTest

import pymongo

from yadm.database import Database
from yadm.documents import Document
from yadm.queryset import QuerySet
from yadm.serialize import from_mongo
from yadm import fields


class BaseDatabaseTest(TestCase):
    skip = False

    def setUp(self):
        if self.skip:
            raise SkipTest

        try:
            self.client = pymongo.MongoClient("localhost", 27017, tz_aware=True)
            self.client.drop_database('test')
        except pymongo.errors.ConnectionFailure:
            self.__class__.skip = True
            raise SkipTest

        self.db = Database(self.client, 'test')


class DatabaseTest(BaseDatabaseTest):
    def test_init(self):
        self.assertIsInstance(self.db.client, pymongo.mongo_client.MongoClient)
        self.assertIsInstance(self.db.db, pymongo.database.Database)
        self.assertEqual(self.db.db.name, 'test')
        self.assertEqual(self.db.name, 'test')

    def test_get_collection(self):
        class TestDoc(Document):
            __collection__ = 'testdocs'

        collection = self.db._get_collection(TestDoc)

        self.assertIsInstance(collection, pymongo.collection.Collection)
        self.assertEqual(collection.name, 'testdocs')

    def test_get_queryset(self):
        class TestDoc(Document):
            __collection__ = 'testdocs'

        queryset = self.db.get_queryset(TestDoc)

        self.assertIsInstance(queryset, QuerySet)
        self.assertIs(queryset._db, self.db)
        self.assertIs(queryset._document_class, TestDoc)


class DatabaseSaveTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class TestDoc(Document):
            __collection__ = 'testdocs'
            i = fields.IntegerField

        self.TestDoc = TestDoc

    def test_insert(self):
        td = self.TestDoc()
        td.i = 13

        self.db.insert(td)

        self.assertEqual(self.db.db.testdocs.find().count(), 1)
        self.assertEqual(self.db.db.testdocs.find()[0]['i'], 13)
        self.assertFalse(td.__fields_changed__)
        self.assertIs(td.__db__, self.db)

    def test_save_new(self):
        td = self.TestDoc()
        td.i = 13

        self.db.save(td)

        self.assertEqual(self.db.db.testdocs.find().count(), 1)
        self.assertEqual(self.db.db.testdocs.find()[0]['i'], 13)
        self.assertFalse(td.__fields_changed__)
        self.assertIs(td.__db__, self.db)

    def test_save(self):
        col = self.db.db.testdocs
        col.insert({'i': 13})

        td = from_mongo(self.TestDoc, col.find_one({'i': 13}))
        td.i = 26

        self.db.save(td)

        self.assertEqual(self.db.db.testdocs.find().count(), 1)
        self.assertEqual(self.db.db.testdocs.find()[0]['i'], 26)
        self.assertFalse(td.__fields_changed__)
        self.assertIs(td.__db__, self.db)


class DatabaseRemoveTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class TestDoc(Document):
            __collection__ = 'testdocs'
            i = fields.IntegerField

        self.TestDoc = TestDoc

    def test_remove(self):
        col = self.db.db.testdocs
        col.insert({'i': 13})

        td = from_mongo(self.TestDoc, col.find_one({'i': 13}))

        self.assertEqual(col.count(), 1)
        self.db.remove(td)
        self.assertEqual(col.count(), 0)



from unittest import TestCase

from yadm import fields
from yadm.documents import Document

from .test_database import BaseDatabaseTest


class PointTest(TestCase):
    def test_init(self):
        point = fields.Point(1, 2)
        self.assertEqual(point.latitude, 1)
        self.assertEqual(point.longitude, 2)

    def test_get(self):
        latitude, longitude = fields.Point(1, 2)
        self.assertEqual(latitude, 1)
        self.assertEqual(longitude, 2)

    def test_to_mongo(self):
        point = fields.Point(1, 2)
        self.assertEqual(
            point.to_mongo(),
            {'type': 'Point', 'coordinates': [1, 2]},
        )

    def test_from_mongo(self):
        point = fields.Point.from_mongo({'type': 'Point', 'coordinates': [1, 2]})
        self.assertIsInstance(point, fields.Point)
        self.assertEqual(point.latitude, 1)
        self.assertEqual(point.longitude, 2)


class MultyPointTest(TestCase):
    def test_init(self):
        mpoint = fields.MultiPoint([fields.Point(1, 2), fields.Point(3, 4)])

        self.assertEqual(len(mpoint._points), 2)
        self.assertEqual(len(mpoint), 2)
        self.assertIsInstance(mpoint[0], fields.Point)

        self.assertEqual(mpoint[0].latitude, 1)
        self.assertEqual(mpoint[0].longitude, 2)
        self.assertEqual(mpoint[1].latitude, 3)
        self.assertEqual(mpoint[1].longitude, 4)

    def test_to_mongo(self):
        mpoint = fields.MultiPoint([fields.Point(1, 2), fields.Point(3, 4)])
        self.assertEqual(
            mpoint.to_mongo(),
            {'type': 'MultiPoint', 'coordinates': [[1, 2], [3, 4]]},
        )

    def test_from_mongo(self):
        mpoint = fields.MultiPoint.from_mongo(
            {'type': 'MultiPoint', 'coordinates': [[1, 2], [3, 4]]}
        )
        self.assertIsInstance(mpoint, fields.MultiPoint)
        self.assertEqual(len(mpoint), 2)
        self.assertIsInstance(mpoint[0], fields.Point)

        self.assertEqual(mpoint[0].latitude, 1)
        self.assertEqual(mpoint[0].longitude, 2)
        self.assertEqual(mpoint[1].latitude, 3)
        self.assertEqual(mpoint[1].longitude, 4)


class ComplexPointTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class TestDoc(Document):
            __collection__ = 'testdoc'
            point = fields.PointField()

        self.TestDoc = TestDoc

    def test_insert(self):
        td = self.TestDoc()
        td.point = fields.Point(1, 2)
        self.db.insert(td)

        raw = self.db.db.testdoc.find_one({'_id': td.id})
        self.assertIn('point', raw)
        self.assertEqual(raw['point'], {'type': 'Point', 'coordinates': [1, 2]})

    def test_get(self):
        _id = self.db.db.testdoc.insert(
            {'point': {'type': 'Point', 'coordinates': [1, 2]}}
        )
        td = self.db(self.TestDoc).with_id(_id)

        self.assertTrue(hasattr(td, 'point'))
        self.assertIsInstance(td.point, fields.Point)
        self.assertEqual(td.point.latitude, 1)
        self.assertEqual(td.point.longitude, 2)



from yadm import fields
from yadm.documents import Document

from .test_database import BaseDatabaseTest


class MapFieldTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class TestDoc(Document):
            __collection__ = 'testdoc'
            map = fields.MapField(fields.IntegerField())

        self.TestDoc = TestDoc

    def test_default(self):
        td = self.TestDoc()
        self.assertIsInstance(td.map, fields.map.Map)
        self.assertFalse(td.map)
        self.assertEqual(len(td.map), 0)
        self.assertEqual(td.map._data, {})
        self.assertEqual(td.map, {})

    def test_get(self):
        _id = self.db.db.testdoc.insert({'map': {'a': 1, 'b': 2, 'c': 3}})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)

        self.assertTrue(td.map)
        self.assertEqual(len(td.map), 3)
        self.assertEqual(td.map._data, {'a': 1, 'b': 2, 'c': 3})
        self.assertEqual(dict(td.map), {'a': 1, 'b': 2, 'c': 3})
        self.assertEqual(td.map, {'a': 1, 'b': 2, 'c': 3})
        self.assertEqual(td.map['b'], 2)

    def test_setattr(self):
        _id = self.db.db.testdoc.insert({'map': {'a': 1, 'b': 2, 'c': 3}})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        td.map['d'] = 4

        self.assertEqual(td.map, {'a': 1, 'b': 2, 'c': 3, 'd': 4})

    def test_setattr_valueerror(self):
        td = self.TestDoc()
        with self.assertRaises(ValueError):
            td.map['d'] = 'not a number'

    def test_setattr_keyerror(self):
        td = self.TestDoc()
        with self.assertRaises(KeyError):
            td.map['not exist']

    def test_setattr_save(self):
        _id = self.db.db.testdoc.insert({'map': {'a': 1, 'b': 2, 'c': 3}})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        td.map['d'] = 4
        self.db.save(td)

        data = self.db.db.testdoc.find_one({'_id': _id})
        self.assertEqual(data['map'], {'a': 1, 'b': 2, 'c': 3, 'd': 4})
        self.assertEqual(td.map, {'a': 1, 'b': 2, 'c': 3, 'd': 4})

    def test_remove(self):
        _id = self.db.db.testdoc.insert({'map': {'a': 1, 'b': 2, 'c': 3}})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        del td.map['b']

        self.assertEqual(td.map, {'a': 1, 'c': 3})

    def test_remove_save(self):
        _id = self.db.db.testdoc.insert({'map': {'a': 1, 'b': 2, 'c': 3}})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        del td.map['b']
        self.db.save(td)

        data = self.db.db.testdoc.find_one({'_id': _id})
        self.assertEqual(data['map'], {'a': 1, 'c': 3})

    def test_set(self):
        _id = self.db.db.testdoc.insert({'map': {'a': 1, 'b': 2, 'c': 3}})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        td.map.set('d', 4)

        self.assertEqual(td.map, {'a': 1, 'b': 2, 'c': 3, 'd': 4})

        data = self.db.db.testdoc.find_one({'_id': _id})
        self.assertEqual(data['map'], {'a': 1, 'b': 2, 'c': 3, 'd': 4})

    def test_set_runtimeerror(self):
        td = self.TestDoc()
        with self.assertRaises(RuntimeError):
            td.map.set('key', 1)

    def test_set_valueerror(self):
        td = self.TestDoc()
        self.db.save(td)
        with self.assertRaises(ValueError):
            td.map.set('key', 'not a number')

    def test_unset(self):
        _id = self.db.db.testdoc.insert({'map': {'a': 1, 'b': 2, 'c': 3}})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        td.map.unset('b')

        self.assertEqual(td.map, {'a': 1, 'c': 3})

        data = self.db.db.testdoc.find_one({'_id': _id})
        self.assertEqual(data['map'], {'a': 1, 'c': 3})



"""
Base classes for build database fields.
"""

from yadm.markers import AttributeNotSet, NoDefault, NotLoaded


class FieldDescriptor(object):
    """ Base desctiptor for fields
    """
    def __init__(self, name, field):
        self.name = name
        self.field = field

    def __repr__(self):
        class_name = type(self).__name__
        document_class_name = type(self.field.document_class).__name__
        return '<{} "{}.{}">'.format(class_name, document_class_name, self.name)

    def __get__(self, instance, owner):
        if instance is None:
            return self.field

        else:
            value = instance.__data__.get(self.name, AttributeNotSet)

            if value is AttributeNotSet:
                raise AttributeError(self.name)

            elif value is NotLoaded:
                value = self.load_deferred(instance)

            value = self.field.from_mongo(instance, value)
            instance.__data__[self.name] = value

            from yadm.documents import DocumentItemMixin

            if isinstance(value, DocumentItemMixin):
                value.__name__ = self.field.name
                value.__parent__ = instance

            return value

    def __set__(self, instance, value):
        if not isinstance(instance, type):
            value = self.field.prepare_value(instance, value)

            if value != instance.__data__.get(self.name):
                instance.__fields_changed__.add(self.name)

            instance.__data__[self.name] = value

        else:
            setattr(instance, self.name, value)

    def __delete__(self, instance):
        if not isinstance(instance, type):
            setattr(instance, self.name, AttributeNotSet)

    def load_deferred(self, instance):
        root = getattr(instance, '__document__', None)

        if root is None:
            qs = instance.__db__.get_queryset(instance.__class__)
            qs = qs.fields(self.name)
            doc = qs.with_id(instance.id)
            value = doc.__data__[self.name]
            value = value if value is not NotLoaded else None
            instance.__data__[self.name] = value

        else:
            qs = instance.__db__.get_queryset(root.__class__)
            proj = [i for i in instance.__path_names__ if isinstance(i, str)]
            qs = qs.fields('.'.join(proj + [self.name]))
            doc = qs.with_id(root.id)
            value = instance.__get_value__(doc).__data__[self.name]
            value = value if value is not NotLoaded else None
            instance.__data__[self.name] = value

        return value


class Field(object):
    """ Base field for all batabase fields

    .. py:attribute:: descriptor_class

        Class of desctiptor for work with field

    .. py:attribute:: default (NoDefault)

        Default value. It is not processed by ``prepare_value``
        in the creating document
    """

    descriptor_class = FieldDescriptor
    default = NoDefault
    name = None
    document_class = None

    def __init__(self, default=NoDefault):
        if default is not NoDefault:
            self.default = self.prepare_value(None, default)

    def __repr__(self):
        class_name = type(self).__name__
        document_class_name = type(self.document_class).__name__
        return '<{} "{}.{}">'.format(class_name, document_class_name, self.name)

    def contribute_to_class(self, document_class, name):
        """ Add field for document_class

        :param MetaDocument document_class: document class for add
        """
        self.name = name
        self.document_class = document_class
        self.document_class.__fields__[name] = self
        setattr(document_class, name, self.descriptor_class(name, self))

    def copy(self):
        """ Return copy of field
        """
        return self.__class__(default=self.default)

    def prepare_value(self, document, value):
        """ The method is called when value is assigned for the attribute

        :param BaseDocument document: document
        :param value: raw value
        :return: prepared value

        It must be accept one argument and return processed (e.g. casted) analog.
        Also it is called for the default value in the creating (not instance!)
        document.
        """
        return value

    def to_mongo(self, document, value):
        return value

    def from_mongo(self, document, value):
        return value



from bson import ObjectId

from yadm.documents import Document
from yadm import fields
from yadm.markers import NotLoaded

from .test_database import BaseDatabaseTest


class QuerySetTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        for n in range(10):
            self.db.db.testdocs.insert({
                'i': n,
                's': 'str({})'.format(n),
            })

        class TestDoc(Document):
            __collection__ = 'testdocs'
            i = fields.IntegerField
            s = fields.StringField

        self.TestDoc = TestDoc
        self.qs = self.db.get_queryset(TestDoc)

    def test_count(self):
        qs = self.qs.find({'i': {'$gte': 6}})
        self.assertEqual(qs.count(), 4)

    def test_len(self):
        qs = self.qs.find({'i': {'$gte': 6}})
        self.assertEqual(len(qs), 4)

    def test_find_one(self):
        td = self.qs.find_one({'i': 7})
        self.assertIsInstance(td, self.TestDoc)
        self.assertTrue(hasattr(td, 'i'))
        self.assertEqual(td.i, 7)

    def test_find(self):
        qs = self.qs.find({'i': {'$gte': 6}})
        self.assertEqual(len([d for d in qs]), 4)
        self.assertEqual({d.i for d in qs}, {6, 7, 8, 9})

    def test_update(self):
        self.qs.find({'i': {'$gte': 6}}).update({'$set': {'s': 'test'}})

        self.assertEqual(self.db.db.testdocs.count(), 10)
        self.assertEqual({d['i'] for d in self.db.db.testdocs.find()}, set(range(10)))
        self.assertEqual(self.db.db.testdocs.find({'s': 'test'}).count(), 4)

        for doc in self.db.db.testdocs.find({'i': {'$lt': 6}}):
            self.assertNotEqual(doc['s'], 'test')
            self.assertTrue(doc['s'].startswith('str('))

        for doc in self.db.db.testdocs.find({'i': {'$gte': 6}}):
            self.assertEqual(doc['s'], 'test')

    def test_update_not_multi(self):
        self.qs.find({'i': {'$gte': 6}}).update({'$set': {'s': 'test'}}, multi=False)

        self.assertEqual(self.db.db.testdocs.count(), 10)
        self.assertEqual({d['i'] for d in self.db.db.testdocs.find()}, set(range(10)))
        self.assertEqual(self.db.db.testdocs.find({'s': 'test'}).count(), 1)

    def test_find_and_modify(self):
        doc = self.qs({'i': 6}).find_and_modify({'$set': {'s': 'test'}})

        self.assertEqual(self.db.db.testdocs.count(), 10)
        self.assertEqual({d['i'] for d in self.db.db.testdocs.find()}, set(range(10)))
        self.assertEqual(self.db.db.testdocs.find({'s': 'test'}).count(), 1)
        self.assertEqual(self.db.db.testdocs.find_one({'i': 6})['s'], 'test')

        self.assertIsInstance(doc, Document)
        self.assertEqual(doc.i, 6)
        self.assertEqual(doc.s, 'str(6)')

    def test_find_and_modify_new(self):
        doc = self.qs({'i': 6}).find_and_modify(
            {'$set': {'s': 'test'}}, new=True)

        self.assertEqual(self.db.db.testdocs.count(), 10)
        self.assertEqual({d['i'] for d in self.db.db.testdocs.find()}, set(range(10)))
        self.assertEqual(self.db.db.testdocs.find({'s': 'test'}).count(), 1)
        self.assertEqual(self.db.db.testdocs.find_one({'i': 6})['s'], 'test')

        self.assertIsInstance(doc, Document)
        self.assertEqual(doc.i, 6)
        self.assertEqual(doc.s, 'test')

    def test_find_and_modify_full_response(self):
        result = self.qs({'i': 6}).find_and_modify(
            {'$set': {'s': 'test'}}, full_response=True)

        self.assertEqual(self.db.db.testdocs.count(), 10)
        self.assertEqual({d['i'] for d in self.db.db.testdocs.find()}, set(range(10)))
        self.assertEqual(self.db.db.testdocs.find({'s': 'test'}).count(), 1)
        self.assertEqual(self.db.db.testdocs.find_one({'i': 6})['s'], 'test')

        self.assertIn('value', result)
        self.assertIsInstance(result['value'], Document)
        self.assertEqual(result['value'].i, 6)
        self.assertEqual(result['value'].s, 'str(6)')

    def test_find_and_modify_full_response_new(self):
        result = self.qs({'i': 6}).find_and_modify(
            {'$set': {'s': 'test'}}, full_response=True, new=True)

        self.assertEqual(self.db.db.testdocs.count(), 10)
        self.assertEqual({d['i'] for d in self.db.db.testdocs.find()}, set(range(10)))
        self.assertEqual(self.db.db.testdocs.find({'s': 'test'}).count(), 1)
        self.assertEqual(self.db.db.testdocs.find_one({'i': 6})['s'], 'test')

        self.assertIn('value', result)
        self.assertIsInstance(result['value'], Document)
        self.assertEqual(result['value'].i, 6)
        self.assertEqual(result['value'].s, 'test')

    def test_find_and_modify_sort(self):
        qs = self.qs({'i': {'$lte': 6, '$gte': 4}}).sort(('s', -1))
        doc = qs.find_and_modify({'$set': {'s': 'test'}})

        self.assertEqual(self.db.db.testdocs.count(), 10)
        self.assertEqual({d['i'] for d in self.db.db.testdocs.find()}, set(range(10)))
        self.assertEqual(self.db.db.testdocs.find({'s': 'test'}).count(), 1)
        self.assertEqual(self.db.db.testdocs.find_one({'i': 6})['s'], 'test')

        self.assertIsInstance(doc, Document)
        self.assertEqual(doc.i, 6)
        self.assertEqual(doc.s, 'str(6)')

    def test_find_and_modify_not_found(self):
        ret = self.qs({'i': 13}).find_and_modify({'$set': {'s': 'test'}})
        self.assertIsNone(ret)

    def test_remove(self):
        self.qs.find({'i': {'$gte': 6}}).remove()
        self.assertEqual(len([d for d in self.qs]), 6)
        self.assertEqual({d.i for d in self.qs}, set(range(6)))

        self.assertEqual(self.db.db.testdocs.count(), 6)
        self.assertEqual({d['i'] for d in self.db.db.testdocs.find()}, set(range(6)))

    def test_sort(self):
        qs = self.qs.find({'i': {'$gte': 6}}).sort(('i', -1))
        self.assertEqual([d.i for d in qs], [9, 8, 7, 6])

    def test_fields(self):
        doc = self.qs.fields('s').find_one({'i': 3})
        self.assertIn('i', doc.__data__)
        self.assertIs(doc.__data__['i'], NotLoaded)
        self.assertEqual(doc.s, 'str(3)')

    def test_fields_all(self):
        doc = self.qs.fields('s').fields_all().find_one({'i': 3})
        self.assertIn('i', doc.__data__)
        self.assertIs(doc.__data__['i'], 3)
        self.assertEqual(doc.s, 'str(3)')
        self.assertEqual(doc.i, 3)

    def test_with_id(self):
        id = self.db.db.testdocs.find_one({'i': 4}, {'_id': True})['_id']
        doc = self.qs.with_id(id)
        self.assertEqual(doc.s, 'str(4)')
        self.assertEqual(doc.i, 4)

    def test_contains(self):
        doc = self.qs.find_one({'i': 0})
        self.assertIn(doc, self.qs)
        self.assertNotIn(doc, self.qs.find({'i': {'$ne': 0}}))

    def test_bulk(self):
        bulk = self.qs.find({'i': {'$gte': 6}}).bulk()
        self.assertIsInstance(bulk, dict)
        self.assertEqual(len(bulk), 4)
        self.assertIsInstance(list(bulk)[0], ObjectId)
        self.assertIsInstance(list(bulk.values())[0], self.TestDoc)
        _id = list(bulk)[0]
        self.assertEqual(bulk[_id].id, _id)
        self.assertEqual({d.i for d in bulk.values()}, {6, 7, 8, 9})



from .simple import StringField


class EmailField(StringField):
    def prepare_value(self, document, value):
        value = super().prepare_value(document, value)

        if value is not None:
            if '@' not in value:
                raise ValueError('"{}" is not email'.format(value))

            return value.lower()



from unittest import TestCase

from bson import ObjectId

from yadm.documents import Document
from yadm import fields


class DocumentsTest(TestCase):
    def setUp(self):
        class TestDoc(Document):
            i = fields.IntegerField
            b = fields.BooleanField

        self.TestDoc = TestDoc

    def test__db(self):
        td = self.TestDoc()
        self.assertIs(td.__db__, None)

    def test_fields(self):
        self.assertEqual(set(self.TestDoc.__fields__), {'_id', 'i', 'b'})

    def test_inheritance_fields(self):
        class InhTestDoc(self.TestDoc):
            d = fields.DecimalField

        self.assertEqual(set(InhTestDoc.__fields__), {'_id', 'i', 'b', 'd'})
        self.assertEqual(set(self.TestDoc.__fields__), {'_id', 'i', 'b'})
        self.assertIsNot(self.TestDoc.__fields__['i'], InhTestDoc.__fields__['i'])
        self.assertIs(self.TestDoc.__fields__['i'].document_class, self.TestDoc)
        self.assertIs(InhTestDoc.__fields__['i'].document_class, InhTestDoc)

    def test_fields_changed(self):
        td = self.TestDoc()

        self.assertNotIn('i', td.__fields_changed__)
        self.assertNotIn('b', td.__fields_changed__)

        td.i = 13
        self.assertIn('i', td.__fields_changed__)
        self.assertNotIn('b', td.__fields_changed__)

        td.b = True
        self.assertIn('i', td.__fields_changed__)
        self.assertIn('b', td.__fields_changed__)

    def test_eq(self):
        doc_a = self.TestDoc()
        doc_a.id = ObjectId()

        doc_b = self.TestDoc()
        doc_b.id = ObjectId()

        doc_c = self.TestDoc()
        doc_c.id = doc_a.id

        self.assertNotEqual(doc_a, doc_b)
        self.assertEqual(doc_a, doc_c)
